@@ -1,95 +1,74 @@
 # sql.R
 # high level functions for sql database access
-# $Id: sql.R,v 0.21 2000/05/23 23:25:26 ml Exp $
 #
 #
 #
 ###########################################
 
-"sqlDrop"<-
-function(channel, sqtable, errors = TRUE)
+sqlDrop <- function(channel, sqtable, errors = TRUE)
 {
-    if (channel < 0)
-        stop("invalid channel")
-    if(missing(sqtable))
-        stop("Missing parameter")
-    tablename<-as.character(substitute(sqtable))
-    if(length(tablename) != 1)
-        stop(paste(tablename,"should be a name"))
-    if(!any(sqlTables(channel)[3] == tablename))
-        stop(paste(tablename,":table not found on channel",channel))
-    query<- paste ("DROP TABLE",tablename)
-    sqlQuery(channel, query, errors = errors)
+    if (channel < 0) stop("invalid channel")
+    if(missing(sqtable)) stop("Missing parameter")
+    tablename <- as.character(substitute(sqtable))
+    odbcTableExists(channel, tablename)
+    sqlQuery(channel, paste ("DROP TABLE", tablename), errors = errors)
 }
 
 
-"sqlFetch" <-
-function (channel, sqtable, ..., colnames = FALSE, rownames = FALSE)
+sqlFetch <-
+    function (channel, sqtable, ..., colnames = FALSE, rownames = FALSE)
 {
     if (channel < 0)
         stop("invalid channel")
     if (missing(sqtable))
         stop("Missing parameter")
     tablename <- as.character(substitute(sqtable))
-    if (length(tablename) != 1)
-        stop(paste(tablename, "should be a name"))
-    switch(odbcCaseFlag(channel),
-	toupper = tablename <- toupper(tablename),
-	tolower = tablename <- tolower(tablename)
-	)
-    if (!any(sqlTables(channel)[3] == tablename))
-        stop(paste(tablename, ":table not found on channel", channel))
-    query <- paste("SELECT * FROM", tablename)
-    ans <- sqlQuery(channel, query, ...)
+    dbname <- odbcTableExists(channel, tablename)
+    ans <- sqlQuery(channel, paste("SELECT * FROM", dbname), ...)
     if (is.logical(colnames) && colnames) {
         colnames(ans) <- as.character(as.matrix(ans[1, ]))
-        ans <- ans[2:nrow(ans), ]
+        ans <- ans[-1, ]
     }
     if (is.logical(rownames) && rownames) {
         rownames(ans) <- as.character(as.matrix(ans[, 1]))
-        ans <- ans[, 2:ncol(ans)]
+        ans <- ans[, -1]
     }
     ans
 }
 
-"sqlClear" <-
-function(channel, sqtable, errors = TRUE)
+sqlClear <- function(channel, sqtable, errors = TRUE)
 {
     if (channel < 0)
         stop("invalid channel")
     if(missing(sqtable))
         stop("Missing parameter")
-    tablename<-as.character(substitute(sqtable))
-    if(length(tablename) != 1)
-        stop(paste(tablename,"should be a name"))
-    if(!any(sqlTables(channel)[3] == tablename))
-        stop(paste(tablename,":table not found on channel",channel))
-    query<- paste ("DELETE FROM",tablename)
-    sqlQuery(channel,query, errors = errors)
+    tablename <- as.character(substitute(sqtable))
+    odbcTableExists(channel, tablename)
+    sqlQuery(channel, paste ("DELETE FROM", tablename), errors = errors)
 }
 
-"sqlCopy" <-
-function(channel, query, destination, destchannel=-1, verbose = TRUE,
-         errors = TRUE)
+sqlCopy <-
+    function(channel, query, destination, destchannel = -1, verbose = TRUE,
+             errors = TRUE)
 {
-    stop("this function does not work yet")
+    .NotYetImplemented()
     if (channel < 0)
         stop("invalid channel")
     if( missing(query) || missing(destination))
         stop("Missing parameter")
     if(destchannel < 0)
         destchannel <- channel
-    tablename<-as.character(substitute(destination))
+    tablename <- as.character(substitute(destination))
     if(length(tablename) != 1)
-        stop(paste(tablename,"should be a name"))
+        stop(paste(tablename, "should be a name"))
     dataset <- sqlQuery(channel, query, errors = errors)
     assign(as.character(substitute(destination)), dataset)
     sqlSave(destchannel, destination, verbose=verbose)
 }
 
-"sqlCopyTable" <-
-function (channel, srctable, desttable, destchannel=-1, verbose = TRUE,
-          errors = TRUE)
+sqlCopyTable <-
+    function (channel, srctable, desttable, destchannel=-1, verbose = TRUE,
+              errors = TRUE)
 {
     if (channel < 0)
         stop("invalid channel")
@@ -99,15 +78,11 @@ function (channel, srctable, desttable, destchannel=-1, verbose = TRUE,
     dtablename <- as.character(substitute(desttable))
     if(length(dtablename) != 1)
         stop(paste(dtablename, "should be a name"))
-    stablename<-as.character(substitute(srctable))
-    if(length(stablename) != 1)
-        stop(paste(stablename, "should be a name"))
-    if(!any(sqlTables(channel)[3]==stablename))
-        stop(paste(stablename,":table not found on channel", channel))
-    query <- sqltablecreate(dtablename, sqlColumns(channel,stablename),
+    stablename <- as.character(substitute(srctable))
+    odbcTableExists(channel, stablename)
+    query <- sqltablecreate(dtablename, sqlColumns(channel, stablename),
                             sqlPrimaryKeys(channel, stablename))
-    if(verbose)
-        print(query)
+    if(verbose) cat("Query: ", query, "\n", sep = "")
     sqlQuery(destchannel, query, errors=errors)
 }
 
@@ -121,9 +96,9 @@ function (channel, srctable, desttable, destchannel=-1, verbose = TRUE,
 
 
 
-"sqlSave" <-
-function(channel, dat, tablename = NULL, append = FALSE, rownames = FALSE,
-         colnames = FALSE, verbose = FALSE, ...)
+sqlSave <-
+    function(channel, dat, tablename = NULL, append = FALSE, rownames = FALSE,
+             colnames = FALSE, verbose = FALSE, oldstyle = FALSE, ...)
 {
     if (channel < 0)
         stop("invalid channel")
@@ -132,51 +107,49 @@ function(channel, dat, tablename = NULL, append = FALSE, rownames = FALSE,
     if(!is.data.frame(dat))
         stop("Should be a dataframe")
     if(is.null(tablename))
-        tablename<-if(length(substitute(dat))==1)
+        tablename <- if(length(substitute(dat)) == 1)
             as.character(substitute(dat))
         else
             as.character(substitute(dat)[[2]])
     if(length(tablename) != 1)
         stop(paste(tablename,"should be a name"))
     switch(odbcCaseFlag(channel),
-           toupper=tablename <- toupper(tablename),
-           tolower=tablename <- tolower(tablename)
+           toupper={tablename <- toupper(tablename)
+                    colnames(dat) <- toupper(colnames(dat))},
+           tolower={tablename <- tolower(tablename)
+                    colnames(dat) <- tolower(colnames(dat))}
            )
-    switch(odbcCaseFlag(channel),
-           toupper=colnames(dat) <- toupper(colnames(dat)),
-           tolower=colnames(dat) <- tolower(colnames(dat))
-           )
-    ## move row labels into data frame
 
-    if(is.logical(rownames) && rownames){
+    ## move row labels into data frame
+    if(is.logical(rownames) && rownames) {
         cbind(row.names(dat),dat)->dat
         names(dat)[1] <- "rownames"
     } else {
-        if(is.character(rownames)){
+        if(is.character(rownames)) {
             cbind(row.names(dat),dat)->dat
             names(dat)[1] <- rownames
         }
     }
 	## ? move col names into dataframe (needed to preserve case)
-    if(is.logical(colnames) && colnames){
+    if(is.logical(colnames) && colnames) {
         ## this is to deal with type conversions
         as.data.frame(rbind(colnames(dat), as.matrix(dat)))->dat
     }
     ## find out if table already exists
-    if(any(sqlTables(channel)[3] == tablename)){
-        if(!append){
+    if(odbcTableExists(channel, tablename, FALSE)) {
+        if(!append) {
             ## zero table, return if no perms
-            query<- paste ("DELETE FROM", tablename)
-            if(verbose)
-                print(query)
+            query <- paste ("DELETE FROM", tablename)
+            if(verbose) cat("Query: ", query, "\n", sep = "")
             sqlQuery(channel, query, errors = TRUE)
         }
-        if(sqlwrite(channel ,tablename, dat, ...) == -1){
+        if(sqlwrite(channel ,tablename, dat, ...) == -1) {
             ##cannot write: try dropping table
-            query <- paste("DROP TABLE", tablename);
-            if(verbose){
-                print(odbcGetErrMsg(channel))
-                print(query)
+            query <- paste("DROP TABLE", tablename)
+            if(verbose) {
+                cat("sqlwrite returned ", odbcGetErrMsg(channel),
+                    "\n", sep = "")
+                cat("Query: ", query, "\n", sep = "")
             }
             sqlQuery(channel, query, errors = TRUE)
         } else { #success
@@ -188,21 +161,57 @@ function(channel, dat, tablename = NULL, append = FALSE, rownames = FALSE,
 #  -	table with invalid columns
 #	No permissions for existing table should have crashed above
 
-		#build a dataframe for create() a la sqlColumns
-    newname <- as.data.frame(cbind(V1=NA, V2=NA, V3=tablename,
-                                   V4=names(dat), V5=12, V6='varchar',
-                                   V7=255, V8=255, V9=NA, V10=NA, V11=1,
-                                   V12=NA))
+# build a dataframe for sqltablecreate() a la sqlColumns
+    nr <- length(dat)
+    NAr <- rep(NA, nr)
+    newname <- data.frame(V1=NAr, V2=NAr, V3=rep(tablename, nr),
+                          V4=names(dat), V5=rep(12,nr), V6=rep("varchar", nr),
+                          V7=rep(255, nr), V8=rep(255, nr), V9=NAr,
+                          V10=NAr, V11=rep(1, nr), V12=NAr)
+    levels(newname[, 6]) <- c("varchar", "real", "integer", "timestamp")
+
+    if(!oldstyle) {
+        typeinfo <- sqlTypeInfo(channel, "all", errors = FALSE)
+        if(is.data.frame(typeinfo)) {
+            ## Now change types as appropriate.
+            types <- sapply(dat, typeof)
+            facs <- sapply(dat, is.factor)
+            isreal <- (types == "double")
+            isint <- (types == "integer") & !facs
+            if(any(isreal)) {
+                realinfo <- sqlTypeInfo(channel, "double")[, 1]
+                if(length(realinfo) > 0) {
+                    if(length(realinfo) > 1) { # more than one match
+                        nm <- match("double", tolower(realinfo))
+                        if(!is.na(nm)) realinfo <- realinfo[nm]
+                    }
+                    newname[isreal, 6] <- levels(newname[, 6])[2] <- realinfo[1]
+                } else isreal[] <- FALSE
+            }
+            if(any(isint)) {
+                intinfo <- sqlTypeInfo(channel, "integer")[, 1]
+                if(length(intinfo) > 0) {
+                    if(length(intinfo) > 1) { # more than one match
+                        nm <- match("integer", tolower(intinfo))
+                        if(!is.na(nm)) intinfo <- intinfo[nm]
+                    }
+                    newname[isint, 6] <- levels(newname[, 6])[3] <- intinfo[1]
+                } else isint[] <- FALSE
+            }
+            newname[isreal | isint, 7] <- 65535
+        } else {
+            warning("creating a table and type info is not available\n-- using varchar columns only\n")
+        }
+    }
     query <- sqltablecreate(tablename, newname, keys=0)
-    if(verbose)
-        print(query)
+    if(verbose) cat("Query: ", query, "\n", sep = "")
     ##last chance:  let it die if fails
     sqlQuery(channel, query, errors = TRUE)
     if(sqlwrite(channel, tablename, dat, ...) < 0) {
         err <- odbcGetErrMsg(channel)
-        print(err)
+        cat(err, "\n", sep="")
         if(err == "Missing column name")
-            print("Check case conversion parameter in odbcConnect")
+            cat("Check case conversion parameter in odbcConnect\n")
         return(-1)
     }
     invisible(1)
@@ -215,8 +224,9 @@ function(channel, dat, tablename = NULL, append = FALSE, rownames = FALSE,
 
 ##############################################
 
-"sqlwrite" <- function (channel, tablename, mydata, test = FALSE, fast = TRUE,
-                        nastring = NULL)
+sqlwrite <-
+    function (channel, tablename, mydata, test = FALSE, fast = TRUE,
+              nastring = NULL)
 {
     coldata <- sqlColumns(channel, tablename)[6]
     ischar <- !is.na(grep("char",
@@ -237,13 +247,13 @@ function(channel, dat, tablename = NULL, append = FALSE, rownames = FALSE,
                            ") VALUES (",
                            paste(data[i, ], collapse = ","),
                            ")")
-            if (verbose) print(query)
+            if(verbose) cat("Query: ", query, "\n", sep = "")
             if (odbcQuery(channel, query) < 0) return(-1)
         }
     } else {
         query <- paste("INSERT INTO", tablename, "(", cnames, ") VALUES (",
                        paste(rep("?", ncol(mydata)), collapse=","), ")")
-        if (verbose) print(query)
+        if(verbose) cat("Query: ", query, "\n", sep = "")
 	coldata <- sqlColumns(channel, tablename)[c(4, 5, 7, 8, 9)]
 	if (is.data.frame(mydata)) mydata <- as.matrix(mydata)
         if(any(is.na(m <- match(colnames, coldata[, 1])))) return(-1)
@@ -261,8 +271,7 @@ function(channel, dat, tablename = NULL, append = FALSE, rownames = FALSE,
 #	NB: some brain dead systems do not support sqlPKs
 ##############################################
 
-"sqltablecreate" <-
-function (tablename, coldata,keys=-1)
+sqltablecreate <- function (tablename, coldata, keys = -1)
 {
     create <- paste("CREATE TABLE", tablename," (")
     j <- nrow(coldata)
@@ -274,96 +283,84 @@ function (tablename, coldata,keys=-1)
         } else {
             null <- "NOT NULL"
         }
-	if(coldata[i, 7] == 65535) { # no size parameter
-		colsize<-" "
-	} else {
-            colsize<-paste("(",coldata[i,7],") ",sep="")
-        }
+	colsize <-
+            if(coldata[i, 7] == 65535) " " else paste("(",coldata[i,7],") ",
+                      sep="")
 
         create <- paste(create,
                         gsub("[^A-Za-z0-9]+", "", as.character(coldata[i,4])),
                         " ", coldata[i,6], colsize, null, sep="")
-        if(!is.numeric(keys)){
+        if(!is.numeric(keys)) {
             if (as.character(keys[[4]]) == as.character(coldata[i,4]))
                 create <- paste(create, " PRIMARY KEY ")
 	}
-        if (i < j) {
-            create <- paste(create, ",")
-        }
+        if (i < j) create <- paste(create, ",")
     }
     create <- paste(create, ")")
     create
 }
+
 ###############################################
 #
 #####  Query Functions
-#	Return a data.frame according to as.is
+#	Return a data frame according to as.is
 #
 ###############################################
 
-"sqlTables" <-
-    function(channel, errors = FALSE, as.is = TRUE)
+sqlTables <- function(channel, errors = FALSE, as.is = TRUE)
 {
     stat <- odbcTables(channel)
     if (stat < 0)
     {
         if(errors) {
-            if(stat == -2)
-                stop("Invalid channel")
-            else
-                return(odbcGetErrMsg(channel))
+            if(stat == -2) stop("Invalid channel")
+            else return(odbcGetErrMsg(channel))
         } else return(-1)
     } else return(sqlGetResults(channel, as.is = as.is))
 }
 
-"sqlColumns" <-
+sqlColumns <-
     function (channel, sqtable, errors = FALSE, as.is = TRUE, special = FALSE)
 {
     if (!is.character(sqtable))
         sqtable <- as.character(substitute(sqtable))
     if (length(sqtable) != 1)
         stop(paste(sqtable, "should be a name"))
-    if (!any(sqlTables(channel)[3] == sqtable)){
-	if(any(tolower(sqlTables(channel)[,3]) == tolower(sqtable))){
-            caseprob <- "\nCheck case parameter in odbcConnect"}
+    if (!odbcTableExists(channel, sqtable, FALSE)) {
+	caseprob <- if(tolower(sqtable) %in% tolower(sqlTables(channel)[,3]))
+            "\nCheck case parameter in odbcConnect"
 	else
-            caseprob <- ""
-        stop(paste(sqtable, ":table not found on channel", channel, caseprob))
+            ""
+        stop(paste(sqtable, ": table not found on channel", channel, caseprob))
     }
-    if(special)
-        stat <- odbcSpecialColumns(channel, sqtable)
-    else
-        stat <- odbcColumns(channel, sqtable)
+    stat <- if(special) odbcSpecialColumns(channel, sqtable)
+    else odbcColumns(channel, sqtable)
     if (stat < 0) {
         if (errors) {
-            if (stat == -2)
-                stop("Invalid channel")
+            if (stat == -2) stop("Invalid channel")
             else return(odbcGetErrMsg(channel))
         } else return(-1)
     } else return(sqlGetResults(channel, as.is = as.is))
 }
 
-"sqlPrimaryKeys" <-
+sqlPrimaryKeys <-
     function(channel, sqtable, as.is = TRUE, errors = FALSE)
 {
     if(!is.character(sqtable))
         sqtable <- as.character(substitute(sqtable))
     if(length(sqtable) != 1)
         stop(paste(sqtable, "should be a name"))
-    if(!any(sqlTables(channel)[3] == sqtable))
-        stop(paste(sqtable, ":table not found on channel", channel))
+    odbcTableExists(channel, sqtable)
     stat <- odbcPrimaryKeys(channel, sqtable)
     if (stat < 0) {
         if(errors) {
-            if(stat == -2)
-                stop("Invalid channel")
-            else
-                return(odbcGetErrMsg(channel))
+            if(stat == -2) stop("Invalid channel")
+            else return(odbcGetErrMsg(channel))
         } else return(-1)
     } else return(sqlGetResults(channel, as.is = as.is))
 }
 
-"sqlQuery" <-
+sqlQuery <-
     function(channel, query, errors = TRUE, ...)
 {
     if(missing(query))
@@ -371,103 +368,24 @@ function (tablename, coldata,keys=-1)
     stat <- odbcQuery(channel, query)
     if (stat < 0)
     {
-        if(errors){
-            if(stat==-2)
-                stop("Invalid channel")
-            else
-                return(odbcGetErrMsg(channel))
+        if(errors) {
+            if(stat==-2) stop("Invalid channel")
+            else return(odbcGetErrMsg(channel))
         } else return(-1)
     } else return(sqlGetResults(channel, errors = errors, ...))
 }
 
-#	See below.  Retained for safekeeping
-#
-"sqlGetResults.old" <-
-    function(channel, as.is= FALSE, errors= FALSE, transposing= FALSE, dec=".")
-{
-    type.convert <-	function(x, na.strings="NA", as.is= FALSEALSE)
-        .Internal(type.convert(x, na.strings,
-                               as.is,dec=options("dec")$dec))
-    cols<-odbcNumCols(channel)$num;
-    if (cols< 0)
-    {
-        if(errors)
-            return("No data")
-        else
-            return (-1)
-    }
-    rows <- odbcNumRows(channel)$num;
-                                        # get all the attribute names
-    attrnames<- c(odbcColData(channel))$names;
-                                        # now get the data
-    cur <- odbcFetchRow(channel);       # get the first row
-    if (cur$stat < 0)                   # no rows
-        return (invisible(0))
-    dbdata <- rbind(cur$data);
-    if(rows == -1){                     # numRows not supported by ODBC library
-                                        # fix from David Middleton
-        while((cur<-odbcFetchRow(channel))$stat !=-1)
-            dbdata <- rbind(dbdata,cur$data);
-        odbcClearError(channel);
-    } else {
-                                        # keep this for better error reporting
-        if (rows > 1)
-        {
-            for (i in c(2:rows))
-            {
-                cur <- odbcFetchRow(channel);
-                if (cur$stat < 0)
-                {
-                    if(errors)
-                        stop(odbcGetErrMsg(channel))
-                    else
-                        return (-1)
-                }
-                                        # bind the data into a matrix
-                dbdata <- rbind(dbdata,cur$data);
-            }
-        }
-    }
-                                        # ? print a warning for truncated data
-    if(errors)
-        if(any(odbcColData(channel)$length>255))
-            print("Warning: a column has been truncated")
-                                        # return the data as data.frame
-    dbdata <- as.data.frame(dbdata);
-                                        # type conversions on columns: pinched from read.table
-    if(is.logical(as.is)) {
-        as.is <- rep(as.is, length=cols)
-    } else if(is.numeric(as.is)) {
-        if(any(as.is < 1 | as.is > cols))
-            stop("invalid numeric as.is expression")
-        i <- rep(FALSE, cols)
-        i[as.is] <- TRUE
-        as.is <- i
-    } else if (length(as.is) != cols)
-        stop(paste("as.is has the wrong length",
-                   length(as.is),"!= cols =", cols))
-    for (i in 1:cols)
-	if (!as.is[i])
-	    dbdata[[i]] <- type.convert(as.character(dbdata[[i]]))
-
-
-                                        # set the names
-    colnames(dbdata) <- attrnames;
-    return(dbdata);
-}
 
 #	New version of sqlGetResults using batch implementation of
 #	odbcFetchRows.  R Version 1+ only
 #
 
-"sqlGetResults" <-
-function (channel, as.is = FALSE,
-          as = c("data frame", "matrix", "transposed matric"),
-          errors = FALSE, max = 0, buffsize = 1000,
-          nullstring = "NA", na.strings = "NA")
+sqlGetResults <-
+    function (channel, as.is = FALSE,
+              as = c("data frame", "matrix", "transposed matric"),
+              errors = FALSE, max = 0, buffsize = 1000,
+              nullstring = "NA", na.strings = "NA")
 {
-    type.convert <- function(x, na.strings = "NA", as.is = FALSE)
-        .Internal(type.convert(x, na.strings, as.is, dec = options("dec")$dec))
     as.df <- function(A, colnames) {
         ## convert transposed char matrix to data frame
         d <- dim(A)
@@ -481,10 +399,7 @@ function (channel, as.is = FALSE,
     as <- match.arg(as)
     cols <- odbcNumCols(channel)$num
     if (cols < 0) {
-        if (errors){
-            return("Nooo data")
-        }
-        else return(-1)
+        if (errors) return("No data") else return(-1)
     }
     attrnames <- odbcColData(channel)$names
     dbdata <- odbcFetchRows(channel,
@@ -514,7 +429,8 @@ function (channel, as.is = FALSE,
             stop(paste("as.is has the wrong length", length(as.is),
                        "!= cols =", cols))
         for (i in 1:cols) if (!as.is[i])
-            data[[i]] <- type.convert(as.character(data[[i]]))
+            data[[i]] <- type.convert(as.character(data[[i]]),
+                                      dec = options("dec")$dec)
     } else {
         data <- dbdata$data
         if(as == "matrix") colnames(data) <- attrnames
@@ -524,8 +440,8 @@ function (channel, as.is = FALSE,
 }
 
 #################################################
-"sqlUpdate" <-
-function(channel, dat, verbose = FALSE, test = FALSE, nastring = NULL)
+sqlUpdate <-
+    function(channel, dat, verbose = FALSE, test = FALSE, nastring = NULL)
 {
     if (channel < 0)
         stop("invalid channel")
@@ -534,16 +450,7 @@ function(channel, dat, verbose = FALSE, test = FALSE, nastring = NULL)
     if(!is.data.frame(dat))
         stop("Should be a dataframe or matrix")
     tablename <- as.character(substitute(dat))
-    if(length(tablename) != 1)
-        stop(paste(tablename,"should be a name"))
-    switch(odbcCaseFlag(channel),
-           toupper = tablename <- toupper(tablename),
-           tolower = tablename <- tolower(tablename)
-           )
-
-    ## find out if table already exists
-    if(!any(sqlTables(channel)[3] == tablename))
-        stop(paste("table",tablename,"does not exist"))
+    odbcTableExists(channel, tablename)
     cnames <- colnames(dat)
     ## match the transform in tablecreate (get rid of inval chars in col names)
     cnames <- gsub("[^A-Za-z0-9]+","",cnames)
@@ -551,8 +458,7 @@ function(channel, dat, verbose = FALSE, test = FALSE, nastring = NULL)
     ## this may or may not include the unique column depending
     ## on whether or not it is a special column.
     coldata <- sqlColumns(channel,tablename)[c(4,5,7,8,9)]
-    if(is.data.frame(dat))
-        dat <- as.matrix(dat)
+    if(is.data.frame(dat)) dat <- as.matrix(dat)
     ## identify the column that is a unique row specifier along with
     ## its descriptor data
     indexcols <- (sqlColumns(channel, tablename, special = TRUE))
@@ -572,23 +478,19 @@ function(channel, dat, verbose = FALSE, test = FALSE, nastring = NULL)
         coldata <- rbind(coldata, indexcols)
     ## check that no columns are present in the df that are not in the table
     indexcols <- as.character(indexcols[ ,1])
-    for (i in cnames){
+    for (i in cnames) {
         if(!any(i == as.character(coldata[ ,1])))
             stop("dataframe column not in database table")
     }
     paramnames <- ""
     query <- paste("UPDATE", tablename, "SET")
-    for (i in 1:length(cnames)){
-        if(any(cnames[i] == indexcols))
-            next
+    for (i in 1:length(cnames)) {
+        if(any(cnames[i] == indexcols)) next
         query <- paste(query,cnames[i], "=?")
-        if(paramnames == "")
-            paramnames <- cnames[i]
-        else
-            paramnames <- c(paramnames,cnames[i])
-        if(i < length(cnames)){
-            if(any(cnames[i+1] == indexcols))
-                next
+        if(paramnames == "") paramnames <- cnames[i]
+        else paramnames <- c(paramnames,cnames[i])
+        if(i < length(cnames)) {
+            if(any(cnames[i+1] == indexcols)) next
             query <- paste(query, ",")
         }
     }
@@ -602,7 +504,7 @@ function(channel, dat, verbose = FALSE, test = FALSE, nastring = NULL)
     ## this next bit of twiddling extracts the descriptor data for each
     ## column and arranges it as columns (easier to parse in C)
     paramdata <- as.character(as.matrix(coldata[coldata[,1] == paramnames[1],]))
-    for (i in 2:length(paramnames)){
+    for (i in 2:length(paramnames)) {
         paramdata <- cbind(paramdata,
                            as.character(as.matrix(coldata[coldata[,1] == paramnames[i],])))
     }
@@ -611,8 +513,48 @@ function(channel, dat, verbose = FALSE, test = FALSE, nastring = NULL)
     ##		data.  This translates to sequential data when passed to C
     ##	Format is colname, datatype,colsize,buff length,decimal digits.
     ##	NB: C routine depends on 5 fields: changes here must be reflected in C
-    if(test | verbose)
-        print(query)
+    if(test | verbose) cat("Query: ", query, "\n", sep = "")
     odbcUpdate(channel, query, dat, paramdata, test = test, verbose = verbose,
                nastring = nastring)
 }
+
+odbcTableExists <- function(channel, tablename, abort=TRUE)
+{
+    if(length(tablename) != 1)
+        stop(paste(tablename, "should be a name"))
+    switch(odbcCaseFlag(channel),
+	toupper = tablename <- toupper(tablename),
+	tolower = tablename <- tolower(tablename)
+	)
+    res <- sqlTables(channel)
+    tables <- if(is.data.frame(res)) res[, 3] else ""
+    # Excel appends a $
+    stables <- sub("\\$$", "", tables)
+    ans <- tablename %in% stables
+    if(abort) {
+        if(!ans)
+            stop(paste(tablename,": table not found on channel", channel))
+        dbname <- tables[match(tablename, stables)]
+        ans <- if(tablename != dbname) paste("[", tablename, "$]", sep="") else tablename
+    }
+    ans
+}
+
+sqlTypeInfo <-
+    function(channel, type = "all", errors = TRUE, as.is = TRUE)
+{
+    ## This depends on the order of symbolic C consts being standard
+    type <- match(type, c("all", "char", "varchar", "real",
+                         "double", "integer", "smallint", "timestamp"),
+                  nomatch=1) - 1
+    stat <- odbcTypeInfo(channel, type)
+    if (stat < 0)
+    {
+        if(errors) {
+            if(stat == -2)
+                stop("Invalid channel")
+            else
+                return(odbcGetErrMsg(channel))
+        } else return(-1)
+    } else return(sqlGetResults(channel, as.is = as.is, errors = errors))
+}
